      PROGRAM Read_AMI
      !====================================================================
      !
      ! Program:
      !   Read_AMI
      !
      ! Object:
      !   To read the 16 channels level1b data of GK2A AMI and 
      !    Convert into Radiance and reflecivity(visible channels) or
      !    temperature(infrared channels)
      !
      ! EXPLANATION:
      !   Level1b data is the digital number of earth observation.
      !   Value ranges corresponds to the number of bits of each channel.
      !
      !   The Radiance value refers to the value of the visible channel for the wavelength
      !    and the infrared channel refers to the value for the wave number.
      !
      !   The visible channel Radiance is converted to Reflectivity.
      !
      !   Infrared channel Radiance is converted to brightness temperature using the
      !     SenPlanck_GK2A module.
      !   
      !   The first two bits of the observation value indicate the data quality flag (DQF)
      !    * DQF
      !      0: Good pixel
      !      1: Conditionally usable pixel (TBD)
      !      2: Out of scan area pixel (TBD)
      !      3: Error pixel (TBD)
      !   
      ! Notice:
      !   You should set 4 variables:
      !   1. Time (yyyymmddhhmn)
      !   2. AMI level 1b Data path
      !   3. Out File path and name (/foo/bar/output_data.bin)
      !   4. Channel number (1~16)
      !
      !   Output:
      !   1. Radiance
      !   2. Reflectivity(visible) / Tb(Infrared)
      !   3. Data Quality Flag
      !
      ! Test Environment
      !   CentOS Linux 7.3.1611
      !   ifort 18.0.1
      !   netcdf-fortran-4.4.4
      !
      ! Conversion Table Version : 3.0 (20190415)
      !
      ! Created by KMA/NMSC in 29 April 2019 
      !
      !====================================================================

      USE SenPlanck_GK2A
 
      IMPLICIT NONE

      REAL,PARAMETER :: ERR=-999.
      INTEGER :: chan, xsize, ysize
      CHARACTER(LEN=256) :: lv1bpath, fname, outfile
      REAL, ALLOCATABLE, DIMENSION(:,:) :: rad, tb_or_ref
      INTEGER(1),ALLOCATABLE,DIMENSION(:,:) :: qc

      REAL:: Start_UTC
      INTEGER:: i, j
      INTEGER:: Year, Month, Day, Hour, Minute,Jday
      INTEGER :: file_p, nfile
      CHARACTER(LEN=256) :: Band_Chan, Res_Chan, Time
      CHARACTER(len=8)   :: chan_string
      INTEGER, DIMENSION(16) :: xydim
      CHARACTER(LEN=5),DIMENSION(16) :: band
      CHARACTER(LEN=3),DIMENSION(16) :: res
       
      DATA xydim / 11000, 11000, 22000, 11000, 5500, 5500, 5500, 5500, &
                    5500,  5500,  5500,  5500, 5500, 5500, 5500, 5500 /

      DATA band / 'vi004', 'vi005', 'vi006', 'vi008', 'nr013', 'nr016', 'sw038', 'wv063', &
                  'wv069', 'wv073', 'ir087', 'ir096', 'ir105', 'ir112', 'ir123', 'ir133' / 

      DATA res / '010', '010', '005', '010', '020', '020', '020', '020', &
                 '020', '020', '020', '020', '020', '020', '020', '020' /

      CALL getarg(1, Time)
      CALL getarg(2, lv1bpath)
      CALL getarg(3, outfile)
      CALL getarg(4, chan_string)
      
      read(chan_string,*) chan

      xsize=xydim(chan)
      ysize=xydim(chan)
      Band_Chan = band(chan)
      Res_Chan = res(chan)

      fname='gk2a_ami_le1b_'//TRIM(Band_Chan)//'_fd'//TRIM(Res_chan)//'ge_'//TRIM(Time)//'.nc'

      ALLOCATE(rad(xsize,ysize))
      ALLOCATE(tb_or_ref(xsize,ysize))
      ALLOCATE(qc(xsize,ysize))

      !Read and Conversion
      IF(chan .ge. 7)THEN

      CALL IR_process (fname, lv1bpath, chan, xsize, ysize, ERR,&
                      rad, tb_or_ref, qc)

      ELSE

      CALL VIS_process (fname, lv1bpath, chan, xsize, ysize,ERR,&
                      rad, tb_or_ref, qc)

      ENDIF


!      print*,rad(2750,2750)
!      print*,tb_or_ref(2750,2750)

      !Save the result
      OPEN(20,File=TRIM(outfile), &
                ACCESS='Direct',FORM='binary',RECL=xsize*ysize*4)
      WRITE(20,REC=1) rad
      WRITE(20,REC=2) tb_or_ref
      !WRITE(20,REC=3) qc

      DEALLOCATE(rad, tb_or_ref, qc)
      CLOSE(20)

      END PROGRAM Read_AMI

!============================================================================!
       SUBROUTINE VIS_process &
                  (Files, FPath, channel, xsize, ysize, real_unavail, &
                   rad_vis, alb_vis,qc_vis)
!============================================================================!
      IMPLICIT NONE

      INTEGER, PARAMETER :: visChannel = 6
      INTEGER, PARAMETER :: nChannel = 16
      INTEGER:: i,j

      INTEGER, PARAMETER :: fp_kind = SELECTED_REAL_KIND( 15 )
      REAL(fp_kind) ::  rad_tmp

      CHARACTER(LEN = 256) :: Filename
      REAL, DIMENSION(visChannel)   :: rwave, rgain, roffset, rrtoa
      
      INTEGER,            INTENT(IN) :: channel
      INTEGER,            INTENT(IN) :: xsize, ysize
      REAL,               INTENT(IN) :: real_unavail
      CHARACTER(LEN = *), INTENT(IN) :: FPath
      CHARACTER(LEN=256), INTENT(IN) :: Files

      INTEGER(1), DIMENSION(xsize,ysize), INTENT(OUT) ::qc_vis
      REAL,       DIMENSION(xsize,ysize), INTENT(OUT) :: alb_vis,rad_vis

      CHARACTER(LEN=5)   :: BNAME(nChannel)       
      CHARACTER(LEN=3)   :: BAND_RES(nChannel)  

      DATA BNAME /'vi004','vi005','vi006','vi008','nr013', 'nr016', & 
                  'sw038','wv063','wv069','wv073','ir087', &
                  'ir096', 'ir105','ir112','ir123','ir133'/

      DATA BAND_RES /'010','010','005','010','020','020', &
                     '020','020','020','020','020', &
                     '020','020','020','020','020' /

!      DATA rrtoa   / 0.001546123, 0.001647508, 0.001909481, 0.003229793, 0.008353981, 0.012562251 /
       DATA rrtoa / 0.001558245, 0.0016595767, 0.001924484, 0.0032723873, 0.0087081313, 0.0129512876 /

      alb_vis(:,:) = real_unavail
      rad_vis(:,:) = real_unavail
      qc_vis(:,:)  = real_unavail

      print*,'VIS Processing..........'

      FileName = trim(FPath)//Files
      CALL ReadRadianceImageData(FileName, channel, xsize, ysize, real_unavail,rad_vis,qc_vis)

      DO j=1,ysize
                DO i=1,xsize
                        IF(rad_vis(i,j) < 0 )  CYCLE
                        alb_vis(i,j) = rad_vis(i,j) * rrtoa(channel)
                ENDDO
      ENDDO

      END SUBROUTINE VIS_process


!============================================================================!
      SUBROUTINE IR_process (Files, FPath, channel, xsize, ysize, real_unavail,&
                              rad_ir, tbb_ir,qc_ir)
!============================================================================!
      USE SenPlanck_GK2A
      
      IMPLICIT NONE
      
      INTEGER, PARAMETER :: nChannel = 16
  
      INTEGER:: i,j

      INTEGER, PARAMETER :: fp_kind = SELECTED_REAL_KIND( 15 )
      REAL(fp_kind) ::  rad_tmp, tbb_tmp
  
      CHARACTER(LEN = 256) :: Filename
  
      CHARACTER(LEN=5),DIMENSION(nChannel) :: BNAME
      CHARACTER(LEN=3),DIMENSION(nChannel) :: BAND_RES
      
      INTEGER,             INTENT(IN) :: channel
      INTEGER,             INTENT(IN) :: xsize, ysize
      REAL,                INTENT(IN)::real_unavail
      CHARACTER(LEN = 256),INTENT(IN) :: Files
      CHARACTER(LEN = *),  INTENT(IN) :: FPath
      
      INTEGER(1), DIMENSION(xsize,ysize), INTENT(OUT) ::  qc_ir 
      REAL,       DIMENSION(xsize,ysize), INTENT(OUT) :: rad_ir, tbb_ir
  
      DATA BNAME /'vi004','vi005','vi006','vi008','nr013', 'nr016', & ! visible 6channel
                  'sw038','wv063','wv069','wv073','ir087', &
                  'ir096', 'ir105','ir112','ir123','ir133'/

      DATA BAND_RES /'010','010','005','010','020','020', &
                     '020','020','020','020','020', &
                     '020','020','020','020','020' /

      tbb_ir(:,:) = real_unavail
      rad_ir(:,:) = real_unavail
      qc_ir(:,:)  = real_unavail

      print*,'IR Processing..........'

      FileName = trim(FPath)//TRIM(Files)
      CALL ReadRadianceImageData(FileName, channel, xsize, ysize, real_unavail, rad_ir, qc_ir)

      DO j=1,ysize
                DO i=1,xsize
  
                        IF(rad_ir(i,j) < 0 )  CYCLE
                        rad_tmp = rad_ir(i,j)
                        tbb_tmp = SenPlanck_rad2temp_wavenum_quad (channel-6, rad_tmp)
                        tbb_ir(i,j) = tbb_tmp

                ENDDO
      ENDDO 
  
      END SUBROUTINE IR_process

     !============================================================================!
     SUBROUTINE ReadRadianceImageData(FilePath, band, xsize, ysize, real_unavail, Radiance, QCData)
     !============================================================================!
        USE NETCDF

        IMPLICIT NONE
        
        INTEGER, PARAMETER :: single = 4
        INTEGER, PARAMETER :: channel = 16 ! 
        INTEGER :: i,j
        INTEGER :: ncid, grpid, varid

        INTEGER,          DIMENSION(channel) :: bandbits
        REAL,             DIMENSION(channel) :: rgain, roffset
        CHARACTER(LEN=5), DIMENSION(channel) :: BNAME 

        REAL,DIMENSION(:,:),ALLOCATABLE :: dataInr
        INTEGER,DIMENSION(:,:),ALLOCATABLE :: dataIn

        CHARACTER(LEN = *), INTENT(IN) :: FilePath
        INTEGER,            INTENT(IN) :: band
        INTEGER,            INTENT(IN) :: xsize, ysize
        REAL,               INTENT(IN) :: real_unavail

        INTEGER(1), DIMENSION(xsize,ysize), INTENT(OUT) :: QCData
        REAL,       DIMENSION(xsize,ysize), INTENT(OUT) :: Radiance
        
!        DATA rgain  /0.323705166578292, 0.303784847259521, 0.130079284310340, &
!                     0.038309440016747, 0.029732408002019, 0.039840638637543, &
!                    -0.00108005676884204,-0.00967241078615188, -0.00739267794415354, &
!                    -0.00880020391196012,-0.0133143225684762, -0.0165399853140115,   &
!                    -0.0184590481221675, -0.0202820152044296, -0.0219888314604759,-0.0229371469467878 /

!        DATA roffset / -6.474090576171870, -6.075683593750000, -5.203155517578120, &
!                       -3.064743041992180, -1.189296722412100, -0.796813964843750, &
!                      17.6945724487304, 39.6104316711425, 60.554573059082, 72.0830001831054, &
!                      109.062217712402, 135.483322143554, 151.208053588867, 166.153015136718,&
!                      180.166610717773, 187.978332519531 /

        DATA rgain / 0.363545805215835, 0.343625485897064, 0.154856294393539, 0.0457241721451282, &
                     0.0346878096461296, 0.0498007982969284, -0.00108296517282724000,             &
                    -0.0108914673328399, -0.00818779878318309, -0.0096982717514038,               &
                    -0.0144806550815701, -0.0178435463458299, -0.0198196955025196,                &
                    -0.0216744858771562, -0.023379972204566, -0.0243037566542625 /

        DATA roffset / -7.27090454101562, -6.87249755859375, -6.19424438476562, -3.65792846679687, &
                       -1.38751220703125, -0.996017456054687, 17.699987411499, 44.1777038574218,   &
                        66.7480773925781, 79.0608520507812, 118.050903320312, 145.464874267578,    &
                        161.580139160156, 176.713439941406, 190.649627685546, 198.224365234375 /
  
        DATA bandbits / 2047, 2047, 4095, 8191, 4095, 2047, 16383, 4095, &
                        8191, 8191, 8191, 8191, 8191, 8191, 8191, 8191 /

        DATA BNAME /'vi004','vi005','vi006','vi008','nr013', 'nr016',  &
                    'sw038','wv063','wv069','wv073','ir087', &
                    'ir096', 'ir105','ir112','ir123','ir133'/

        ALLOCATE(dataInr(xsize,ysize))
        ALLOCATE(dataIn(xsize,ysize))

        CALL check(nf90_open(path = FilePath,mode = NF90_NOWRITE, ncid = ncid))
        CALL check(nf90_inq_varid(ncid, "image_pixel_values", varid))
        CALL check(nf90_get_var(ncid, varid, dataInr))

        dataIn = int(dataInr)

            DO j = 1, ysize
                    DO i = 1, xsize
                                 QCData(i,j)   = ISHFT(dataIn(i,j),-14)
                                 Radiance(i,j) = IAND(dataIn(i,j),bandbits(band))
 
                                 !! if you need to 'flag process'
                                 !if (QCData(i,j).eq.0) Radiance(i,j)=IAND(dataIn(i,j),bandbits(band))
                                 !if (QCData(i,j).eq.1) Radiance(i,j)=real_unavail
                                 !if (QCData(i,j).eq.2) Radiance(i,j)=real_unavail
                                 !if (QCData(i,j).eq.3) Radiance(i,j)=real_unavail
                                 
                                 if (Radiance(i,j).ne.0) then
                                Radiance(i,j)=(Radiance(i,j)*rgain(band))+roffset(band)
                                endif
                     ENDDO
                END DO
!        print*,DataIn(2750,2750)

        DEALLOCATE(DataIn)

     END SUBROUTINE ReadRadianceImageData

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
 SUBROUTINE check(status)
     USE NETCDF
          integer, intent(in) :: status
        
          if (status /= nf90_noerr) then
              print*, trim(nf90_strerror(status))
              stop "Stopped"
          end if
 END SUBROUTINE check

